#include "symtab.h"


inline bool bit_test(unsigned char * bit_Map, int codepoint) {
	return (bit_Map[codepoint/8] >> (7 - codepoint % 8)) & 1;
}


bool is_XML10_NameStrt_codepoint(int codepoint) {
	switch (codepoint >> 12) {
		case 0: return bit_test(NameStrt_XML10_0000_11FF, codepoint);
		case 1: if (codepoint <= 0x11FF)
				return bit_test(NameStrt_XML10_0000_11FF, codepoint);
			else if (codepoint < 0x1E00) return false;
			else return bit_test(NameStrt_XML10_1E00_1FFF, codepoint & 0x1FF);
		case 2:	if (codepoint > 0x2182) return false;
			else return bit_test(NameStrt_XML10_2000_21FF, codepoint & 0x1FF);
		case 3:	if (codepoint > 0x312C) return false;
			else return bit_test(NameStrt_XML10_3000_31FF, codepoint & 0x1FF);
		case 4: return codepoint >= 0x4E00;
		case 5: case 6: case 7: case 8: return true;
		case 9: return codepoint <= 0x9FA5;
		case 0xA: return codepoint >= 0xAC00;
		case 0xB: case 0xC: return true;
		case 0xD: return codepoint <= 0xD7A3;
		default: return false;
	}
}

bool is_XML10_NameChar_codepoint(int codepoint) {
	switch (codepoint >> 12) {
		case 0: return bit_test(NameChar_XML10_0000_11FF, codepoint);
		case 1: if (codepoint <= 0x11FF)
				return bit_test(NameChar_XML10_0000_11FF, codepoint);
			else if (codepoint < 0x1E00) return false;
			else return bit_test(NameStrt_XML10_1E00_1FFF, codepoint & 0x1FF);
		case 2:	if (codepoint > 0x2182) return false;
			else return bit_test(NameChar_XML10_2000_21FF, codepoint & 0x1FF);
		case 3:	if (codepoint > 0x312C) return false;
			else return bit_test(NameChar_XML10_3000_31FF, codepoint & 0x1FF);
		case 4: return codepoint >= 0x4E00;
		case 5: case 6: case 7: case 8: return true;
		case 9: return codepoint <= 0x9FA5;
		case 0xA:	return codepoint >= 0xAC00;
		case 0xB: case 0xC: return true;
		case 0xD: return codepoint <= 0xD7A3;
		default: return false;
	}
}

bool is_XML11_NameStrt_codepoint(int codepoint) {
	if (likely(codepoint) <= 0x03FF) return bit_test(NameStrt_XML11_0000_03FF, codepoint);
	else switch (codepoint >> 12) {
		case 0: case 1: return true;
		case 2:	if (codepoint >= 0x2070)
				if (codepoint <= 0x218F) return true;
				else return (codepoint >= 0x2C00) & (codepoint <= 0x2FEF);
			else return (codepoint >= 0x200C) & (codepoint <= 0x200D);
		case 3:	return codepoint >= 0x3001;
		case 4: case 5: case 6: case 7: case 8: case 9: case 0xA: case 0xB: case 0xC: return true;
		case 0xD: return codepoint <= 0xD7FF;
		case 0xE: return false;
		case 0xF: if (codepoint <= 0xFDCF) return codepoint >= 0xF900;
			  else return (codepoint >= 0xFDF0) & (codepoint <= 0xFFFD);
		default: return codepoint <= 0xEFFFF;
	}
}

bool is_XML11_NameChar_codepoint(int codepoint) {
	if (likely(codepoint) <= 0x03FF) return bit_test(NameChar_XML11_0000_03FF, codepoint);
	else switch (codepoint >> 12) {
		case 0: case 1: return true;
		case 2:	if (codepoint >= 0x2070)
				if (codepoint <= 0x218F) return true;
				else return (codepoint >= 0x2C00) & (codepoint <= 0x2FEF);
			else if (codepoint <= 0x200D) return codepoint >= 0x200C;
			else return (codepoint == 0x203F) | (codepoint == 0x2040);
		case 3:	return codepoint >= 0x3001;
		case 4: case 5: case 6: case 7: case 8: case 9: case 0xA: case 0xB: case 0xC: return true;
		case 0xD: return codepoint <= 0xD7FF;
		case 0xE: return false;
		case 0xF: if (codepoint <= 0xFDCF) return codepoint >= 0xF900;
			  else return (codepoint >= 0xFDF0) & (codepoint <= 0xFFFD);
		default: return codepoint <= 0xEFFFF;
	}
}

inline int XML_10_UTF8_NameStrt_bytes (unsigned char bytes[]) {
	if (bytes[0] <= 0x7F) {
		if (bit_test(NameStrt_XML10_0000_11FF, (int) bytes[0])) return 1;
		else return 0;
	}
	else if (bytes[0] <= 0xDF) {
		int codepoint = ((bytes[0] & 0x3F) << 6) | (bytes[1] & 0x3F);
		if (bit_test(NameStrt_XML10_0000_11FF, codepoint)) return 2;
		else return 0;
	}
	else if (bytes[0] <= 0xEF) {
		int codepoint = ((bytes[0] & 0x0F) << 12)| ((bytes[1] & 0x3F) << 6) | (bytes[2] & 0x3F);
		return is_XML10_NameStrt_codepoint(codepoint) ? 3 : 0;
	}
	else return 0;
}

inline int XML_10_UTF8_NameChar_bytes (unsigned char bytes[]) {
	if (bytes[0] <= 0x7F) {
		if (bit_test(NameChar_XML10_0000_11FF, (int) bytes[0])) return 1;
		else return 0;
	}
	else if (bytes[0] <= 0xDF) {
		int codepoint = ((bytes[0] & 0x3F) << 6) | (bytes[1] & 0x3F);
		if (bit_test(NameChar_XML10_0000_11FF, codepoint)) return 2;
		else return 0;
	}
	else if (bytes[0] <= 0xEF) {
		int codepoint = ((bytes[0] & 0x0F) << 12)| ((bytes[1] & 0x3F) << 6) | (bytes[2] & 0x3F);
		return is_XML10_NameChar_codepoint(codepoint) ? 3 : 0;
	}
	else return 0;
}

inline int XML_11_UTF8_NameStrt_bytes (unsigned char bytes[]) {
	if (bytes[0] <= 0x7F) {
		if (bit_test(NameStrt_XML11_0000_03FF, (int) bytes[0])) return 1;
		else return 0;
	}
	else if (bytes[0] <= 0xDF) {
		int codepoint = ((bytes[0] & 0x3F) << 6) | (bytes[1] & 0x3F);
		return is_XML11_NameStrt_codepoint(codepoint) ? 2 : 0;
	}
	else if (bytes[0] <= 0xEF) {
		int codepoint = ((bytes[0] & 0x0F) << 12)| ((bytes[1] & 0x3F) << 6) | (bytes[2] & 0x3F);
		return is_XML11_NameStrt_codepoint(codepoint) ? 3 : 0;
	}
	else {
		int codepoint = ((bytes[0] & 0x0F) << 18)| ((bytes[1] & 0x3F) << 12) |
				((bytes[2] & 0x3F) << 6) | (bytes[3] & 0x3F);
		return is_XML11_NameStrt_codepoint(codepoint) ? 4 : 0;
	}
}

inline int XML_11_UTF8_NameChar_bytes (unsigned char bytes[]) {
	if (bytes[0] <= 0x7F) {
		if (bit_test(NameChar_XML11_0000_03FF, (int) bytes[0])) return 1;
		else return 0;
	}
	else if (bytes[0] <= 0xDF) {
		int codepoint = ((bytes[0] & 0x3F) << 6) | (bytes[1] & 0x3F);
		return is_XML11_NameChar_codepoint(codepoint) ? 2 : 0;
	}
	else if (bytes[0] <= 0xEF) {
		int codepoint = ((bytes[0] & 0x0F) << 12)| ((bytes[1] & 0x3F) << 6) | (bytes[2] & 0x3F);
		return is_XML11_NameChar_codepoint(codepoint) ? 3 : 0;
	}
	else {
		int codepoint = ((bytes[0] & 0x0F) << 18)| ((bytes[1] & 0x3F) << 12) |
				((bytes[2] & 0x3F) << 6) | (bytes[3] & 0x3F);
		return is_XML11_NameChar_codepoint(codepoint) ? 4 : 0;
	}
}

bool is_XML10_UTF8_Name(char protoname[], int lgth) {
	int valid_bytes = XML_10_UTF8_NameStrt_bytes((unsigned char *) protoname);
	int pos = valid_bytes;
	while ((valid_bytes > 0) & (pos < lgth)) {
		valid_bytes = XML_10_UTF8_NameChar_bytes((unsigned char *) &protoname[pos]);
		pos += valid_bytes;

	}
	/* Success requires that every byte sequence processed be valid
           and that the total lgth processed be exactly that provided on
           input. */

	return (valid_bytes > 0) & (pos == lgth);
}

bool is_XML11_UTF8_Name(char protoname[], int lgth) {
	int valid_bytes = XML_11_UTF8_NameStrt_bytes((unsigned char *) protoname);
	int pos = valid_bytes;
	while ((valid_bytes > 0) & (pos < lgth)) {
		valid_bytes = XML_11_UTF8_NameChar_bytes((unsigned char *) &protoname[pos]);
		pos += valid_bytes;
	}
	/* Success requires that every byte sequence processed be valid
           and that the total lgth processed be exactly that provided on
           input. */
	return (valid_bytes > 0) & (pos == lgth);
}

bool is_XML10_UTF8_Nmtoken(char prototoken[], int lgth) {
	int valid_bytes = XML_10_UTF8_NameChar_bytes((unsigned char *) prototoken);
	int pos = valid_bytes;
	while ((valid_bytes > 0) & (pos < lgth)) {
		valid_bytes = XML_10_UTF8_NameChar_bytes((unsigned char *) &prototoken[pos]);
		pos += valid_bytes;

	}
	/* Success requires that every byte sequence processed be valid
           and that the total lgth processed be exactly that provided on
           input. */

	return (valid_bytes > 0) & (pos == lgth);
}

bool is_XML11_UTF8_Nmtoken(char prototoken[], int lgth) {
	int valid_bytes = XML_11_UTF8_NameChar_bytes((unsigned char *) prototoken);
	int pos = valid_bytes;
	while ((valid_bytes > 0) & (pos < lgth)) {
		valid_bytes = XML_11_UTF8_NameChar_bytes((unsigned char *) &prototoken[pos]);
		pos += valid_bytes;
	}
	/* Success requires that every byte sequence processed be valid
           and that the total lgth processed be exactly that provided on
           input. */
	return (valid_bytes > 0) & (pos == lgth);
}

int Symbol_Table::Insert_Name(const char * name, int lgth) {
//	char * s = copy_name(name,lgth);
	char * s = pool.Insert(name,lgth);
	UTF8NameMap[s]=++(globalNameCount);
	Name_Data name_data;
	name_data.name_string = s;
	name_data.lgth = lgth;
	UTF8NameTable.push_back(name_data);
	return globalNameCount;
}


inline bool Verify_ASCII(char * name_ptr, int name_lgth) {
	/* To verify that a name is ASCII, ensure that the high bit
	   of each byte is 0.  A SIMD compare can verify this for
	   up to sizeof(BytePack) bytes.  For less than 16 bytes,
	   first shift out bytes beyond the name length.  For more
	   than 16 bytes, form the logical "or" of the successive byte
	   packs together so that a high 1 bit in any byte is preserved
	   for the final SIMD test. */
	BytePack b = sisd_load_unaligned((BytePack *) name_ptr);
	if (name_lgth <= sizeof(BytePack)) {
		/* Clear bytes beyond the length of the name. */
		b = sisd_sfl(b, sisd_from_int(8 * (sizeof(BytePack) - name_lgth)));
	}
	else {
		int offset = name_lgth % sizeof(BytePack);
		for (int i = offset; i < name_lgth; i += sizeof(BytePack)) {
			b = simd_or(sisd_load_unaligned((BytePack *) &name_ptr[i]),b);
		}
	}
#ifdef TEMPLATED_SIMD_LIB
	return !simd_any_sign_bit<8>(b);
#endif
#ifndef TEMPLATED_SIMD_LIB
	return !simd_any_sign_bit_8(b);
#endif
}


/* ASCII_LookupOrInsert determines the nameID for any ASCII name
from the global name table, inserting the name and allocating
a nameID if necessary.  If the name is non-ASCII, 0 is returned. */

inline int Symbol_Table::ASCII_Lookup_or_Insert_Name(char * name_ptr, int name_lgth) {

	if (Verify_ASCII(name_ptr, name_lgth)) {
		return UTF8_Lookup_or_Insert_Name(name_ptr, name_lgth);
	}
	return 0;
}



int Symbol_Table::UTF8_Lookup_or_Insert_Name(char * name, int lgth) {

	char delim = name[lgth];
	name[lgth] = '\0';
	int nameID = UTF8NameMap[name];
	name[lgth] = delim;

	if(nameID == 0){
	#if (not defined(OMISSION)) or (OMISSION != NAME_VALIDATION)
#ifdef EDITION5
		if (!is_XML11_UTF8_Name(name,lgth))  {
			ShowSyntaxError(NT_Name);
			exit(-1);
		}
#endif
#ifndef EDITION5
		if (!is_XML10_UTF8_Name(name,lgth))  {
			if (version == XML_1_1) {
				if (!is_XML11_UTF8_Name(name,lgth))  {
					ShowSyntaxError(NT_Name);
					exit(-1);
				}
			}
			else {
				ShowSyntaxError(NT_Name);
				exit(-1);
			}
		}
#endif
	#endif
//		char * s = copy_name(name,lgth);
		char * s = pool.Insert(name,lgth);
		UTF8NameMap[s]=++(globalNameCount);
		nameID = globalNameCount;
		Name_Data name_data;
		name_data.name_string = s;
		name_data.lgth = lgth;
		UTF8NameTable.push_back(name_data);
//		UTF8NameTable.push_back(s);
	}
	return nameID;
}

//char * Symbol_Table::Get_UTF8_name(int nameID) {
//	return 	UTF8NameTable[nameID];
//}

char * Symbol_Table::Get_UTF8_name(int nameID) {
	return 	UTF8NameTable[nameID].name_string;
}

int Symbol_Table::Get_UTF8_lgth(int nameID) {
	return 	UTF8NameTable[nameID].lgth;
}

inline int Symbol_Table::ASCII_Lookup_or_Insert_Nmtoken(char * nmtoken_ptr, int name_lgth) {

	if (Verify_ASCII(nmtoken_ptr, name_lgth)) {
		return UTF8_Lookup_or_Insert_Nmtoken(nmtoken_ptr, name_lgth);
	}
	return 0;
}


int Symbol_Table::UTF8_Lookup_or_Insert_Nmtoken(char * nmtoken, int lgth) {

	char delim = nmtoken[lgth];
	nmtoken[lgth] = '\0';
	int nmtokenID = UTF8NmtokenMap[nmtoken];
	nmtoken[lgth] = delim;

	if(nmtokenID == 0){
	#if (not defined(OMISSION)) or (OMISSION != NAME_VALIDATION)
#ifdef EDITION5
		if (!is_XML11_UTF8_Nmtoken(nmtoken,lgth))  {
			ShowSyntaxError(NT_Nmtoken);
			exit(-1);
		}
#endif
#ifndef EDITION5
		if (!is_XML10_UTF8_Nmtoken(nmtoken,lgth))  {
			if (version == XML_1_1) {
				if (!is_XML11_UTF8_Nmtoken(nmtoken,lgth))  {
					ShowSyntaxError(NT_Nmtoken);
					exit(-1);
				}
			}
			else {
				ShowSyntaxError(NT_Nmtoken);
				exit(-1);
			}
		}
#endif
	#endif
//		char * s = copy_name(name,lgth);
		char * s = pool.Insert(nmtoken,lgth);
		UTF8NmtokenMap[s]=++(globalNmtokenCount);
		nmtokenID = globalNmtokenCount;
		Name_Data nmtoken_data;
		nmtoken_data.name_string = s;
		nmtoken_data.lgth = lgth;
		UTF8NmtokenTable.push_back(nmtoken_data);
//		UTF8NameTable.push_back(s);
	}
	return nmtokenID;
}




char * Symbol_Table::Get_UTF8_nmtoken(int nmtokenID) {
	return 	UTF8NmtokenTable[nmtokenID].name_string;
}

int Symbol_Table::Get_UTF8_nmtoken_lgth(int nmtokenID) {
	return 	UTF8NmtokenTable[nmtokenID].lgth;
}

char * Symbol_Table::ReserveSymbolSpace(int u8_lgth) {
	reserved = new char[u8_lgth+1];
	reserved_lgth = u8_lgth;
	return reserved;
}

int Symbol_Table::LookupOrInsertReserved(){
	int nameID = UTF8NameMap[reserved];
	if(nameID == 0){
	#if (not defined(OMISSION)) or (OMISSION != NAME_VALIDATION)
		if (!is_XML10_UTF8_Name(reserved,reserved_lgth))  {
			ShowSyntaxError(NT_Name);
			exit(-1);
		}
	#endif
		UTF8NameMap[reserved]=++(globalNameCount);
		nameID = globalNameCount;
		Name_Data name_data;
		name_data.name_string = reserved;
		name_data.lgth = reserved_lgth;
		UTF8NameTable.push_back(name_data);
//		UTF8NameTable.push_back(s);
	}
	else {
		delete [] reserved;
	}
	return nameID;
}

int Symbol_Table::LookupOrInsertReserved_nmtoken(){
	int nmtokenID = UTF8NmtokenMap[reserved];
	if(nmtokenID == 0){
	#if (not defined(OMISSION)) or (OMISSION != NAME_VALIDATION)
		if (!is_XML10_UTF8_Nmtoken(reserved,reserved_lgth))  {
			ShowSyntaxError(NT_Nmtoken);
			exit(-1);
		}
	#endif
		UTF8NmtokenMap[reserved]=++(globalNmtokenCount);
		nmtokenID = globalNmtokenCount;
		Name_Data nmtoken_data;
		nmtoken_data.name_string = reserved;
		nmtoken_data.lgth = reserved_lgth;
		UTF8NmtokenTable.push_back(nmtoken_data);
	}
	else {
		delete [] reserved;
	}
	return nmtokenID;
}

Symbol_Table::Symbol_Table(){
	globalNameCount = 0;
	Name_Data name_data;
	name_data.name_string = NULL;
	name_data.lgth = 0;
	UTF8NameTable.push_back(name_data);
//	UTF8NameTable.push_back(NULL);
/*	for (int i = 0; i < 5; i++) {
		UTF8NameMap[predefined[i]] = ++(globalNameCount);
		Name_Data name_data;
		name_data.name_string = predefined[i];
		name_data.lgth = strlen(predefined[i]);
		UTF8NameTable.push_back(name_data);
		printf("predefined name: %s, global name count: %d\n",predefined[i],globalNameCount);
//		UTF8NameTable.push_back(predefined[i]);
	}*/
}




